import datetime
import logging
from typing import Type, List

from pydantic_db_backend.backend import Backend
from pydantic_db_backend.contexts.custom_aggregation import custom_aggregation_provider
from pydantic_db_backend.contexts.pagination_parameter import (
    pagination_parameter_provider,
)
from pydantic_db_backend.pagination import pagination_parameter_resolve
from pydantic_db_backend_common.exceptions import RevisionConflict
from pydantic_db_backend_common.pydantic import (
    PaginationParameterModel,
    FindResultModel,
    CustomAggregationModel,
)
from pydantic_db_backend_common.utils import utcnow, Undefined

from eventix.exceptions import NoTaskFoundForUniqueKey, WrongTaskStatus, TaskUniqueKeyNotUnique
from eventix.functions.features import backend_features
from eventix.functions.relay import RelayManager
from eventix.pydantic.task import TEventixTask, EventixTaskStatusEnum

# post
# already exists
# still scheduled -> update
# not scheduled --> post new.
# scheduled -> update
#
#   on update:
#       get task
#       scheck scheduled, if not post
#       update task
#
#       if worker grabs , version should conflict
#       on conflict : try again

log = logging.getLogger(__name__)


def task_post(task: TEventixTask) -> TEventixTask:

    # try to relay
    r = RelayManager.try_relay(task)
    if r is not None:
        return r  # relay successful

    # no relay

    is_unique = task.unique_key is not None
    client = Backend.client()

    if not is_unique:
        # not unique , just try to save. If exists, raise error
        # noinspection PyTypeChecker
        return client.post_instance(task)

    # has unique_key

    while True:

        # noinspection PyTypeChecker
        with pagination_parameter_provider(
            PaginationParameterModel(filter={"unique_key": task.unique_key,"namespace":task.namespace})
        ):
            existing_tasks = client.get_instances(TEventixTask)
        next_scheduled_task = next(
            filter(lambda t: t.status in (EventixTaskStatusEnum.scheduled, EventixTaskStatusEnum.retry), existing_tasks.data),
            None,
        )

        if next_scheduled_task is None:
            # no existing ones that are only scheduled, we have to post
            # noinspection PyTypeChecker
            return client.post_instance(task)

        #   update:
        #       get task
        #       scheck scheduled, if not post
        #       update task
        #
        #       if worker grabs , version should conflict
        #       on conflict: try again

        next_scheduled_task.unique_update_from(task)
        try:
            updated_task = client.put_instance(next_scheduled_task)
            # noinspection PyTypeChecker
            log.debug(f"updated task {updated_task.uid}")
            # noinspection PyTypeChecker
            return updated_task  # update worked
        except RevisionConflict:
            continue  # try again.


def task_clean_expired_workers():
    client = Backend.client()
    with pagination_parameter_provider(
        PaginationParameterModel(
            skip=0,
            limit=1,
            filter=dict(worker_expires={"$ne": None, "$lt": utcnow()}),
            sort="priority,eta",
        )
    ):
        while True:
            existing_task: TEventixTask | None = next(
                iter(client.get_instances(model=TEventixTask).data), None
            )

            # repeat until we were able to take something or nothing is left.
            if existing_task is None:
                break

            existing_task.status = EventixTaskStatusEnum.scheduled
            existing_task.worker_id = None
            existing_task.worker_expires = None

            try:
                client.put_instance(existing_task)
                log.info(f"Released task {existing_task.uid}")
                # noinspection PyTypeChecker
            except RevisionConflict:
                continue


def task_clean_expired_tasks():
    client = Backend.client()
    pagination_parameter = PaginationParameterModel(
        skip=0, limit=100, filter=dict(expires={"$ne": None, "$lt": utcnow()})
    )

    while True:
        existing_uids = client.get_uids(TEventixTask, pagination_parameter)

        # repeat until we were able to take something or nothing is left.
        if len(existing_uids.data) == 0:
            break

        for uid in existing_uids.data:
            client.delete_uid(TEventixTask, uid)
            log.info(f"Removed expired task {uid}")


def task_next_scheduled(
    worker_id: str, namespace: str, expires: int = 300
) -> TEventixTask | None:
    log.debug(f"[{worker_id}] Worker getting next scheduled task...")
    client = Backend.client()

    # looking up possible tasks in right order
    # take first one
    # try to set worker_id and expiration

    eta = utcnow()  # eta has to be now or in the past

    pagination_parameter = PaginationParameterModel(
        filter=(
            dict(
                namespace=namespace,  # namespace has to match
                worker_id=None,  # no worker assigned
                status={"$in": [EventixTaskStatusEnum.scheduled, EventixTaskStatusEnum.retry]},
                eta={"$lte": eta},
            )
        ),
        limit=1,
        sort="priority,eta",
    )

    while True:  # repeat until we were able to take something or nothing is left.

        existing_task = next(
            iter(client.get_instances(TEventixTask, pagination_parameter).data), None
        )

        if existing_task is None:
            return None  # no task left

        existing_task.status = "processing"
        existing_task.worker_id = worker_id
        existing_task.worker_expires = utcnow() + datetime.timedelta(seconds=expires)
        log.debug(
            f"task_next_scheduled: existing task revision: {existing_task.revision}"
        )
        try:
            # noinspection PyTypeChecker
            t: TEventixTask = client.put_instance(existing_task)
            return t
        except RevisionConflict:
            continue


def tasks_by_status(
    status: EventixTaskStatusEnum | None = None,
    namespace: str | None = None,
    pagination_parameter: PaginationParameterModel | None | Type[Undefined] = Undefined,
) -> FindResultModel:
    client = Backend.client()
    pagination_parameter = pagination_parameter_resolve(pagination_parameter)

    filter = {}
    if status is not None:
        filter |= {"status": status}

    if namespace is not None:
        filter |= {"namespace": namespace}

    if len(filter) != 0:
        pagination_parameter.filter = (
            filter
            if pagination_parameter.filter is None
            else pagination_parameter.filter | filter
        )

    tasks = client.get_instances(TEventixTask, pagination_parameter)
    return tasks


@backend_features(features=["find_extend_pipeline"])
def tasks_by_task(
    task: str = None,
    namespace: str | None = None,
    pagination_parameter: PaginationParameterModel | None | Type[Undefined] = Undefined,
) -> FindResultModel:
    client = Backend.client()
    pagination_parameter = pagination_parameter_resolve(pagination_parameter)

    filter = {}
    filter |= {"task": task}

    if namespace is not None:
        filter |= {"namespace": namespace}

    pagination_parameter.sort = "status_id,-eta"

    if len(filter) != 0:
        pagination_parameter.filter = (
            filter
            if pagination_parameter.filter is None
            else pagination_parameter.filter | filter
        )

    with custom_aggregation_provider(
        CustomAggregationModel(
            before_tail=[
                {
                    "$addFields": {
                        "status_id": {
                            "$indexOfArray": [
                                [EventixTaskStatusEnum.scheduled, EventixTaskStatusEnum.processing, EventixTaskStatusEnum.retry, EventixTaskStatusEnum.error, EventixTaskStatusEnum.done],
                                "$status",
                            ]
                        }
                    }
                }
            ]
        )
    ):
        tasks = client.get_instances(TEventixTask, pagination_parameter)
    return tasks


def task_by_unique_key(unique_key: str,namespace:str=None, stati:List[EventixTaskStatusEnum]=None) -> TEventixTask:
    filter_params={"unique_key": unique_key}
    if namespace is not None:
        filter_params["namespace"] = namespace

    if stati is None:
        stati = [EventixTaskStatusEnum.scheduled.value, EventixTaskStatusEnum.retry.value]

    existing_tasks = Backend.client().get_instances(
        TEventixTask,
        PaginationParameterModel(limit=10, filter=filter_params),
    )

    filtered = list(filter(lambda t: t.status in stati, existing_tasks.data))
    if len(filtered)==0:
        raise NoTaskFoundForUniqueKey(unique_key=unique_key)

    if len(filtered) >1 and namespace is None:
        raise TaskUniqueKeyNotUnique(unique_key)
    return filtered[0]


def task_reschedule(uid: str, eta: datetime.datetime | None = None) -> TEventixTask:
    # noinspection PyTypeChecker
    t: TEventixTask = Backend.client().get_instance(TEventixTask, uid)

    if t.status not in ["error", "retry"]:
        raise WrongTaskStatus(uid, t.status)

    if eta is None:
        eta = utcnow()

    t.status = EventixTaskStatusEnum.scheduled
    t.eta = eta
    t.worker_id = None
    Backend.client().put_instance(t)
    return t


def tasks_dump(
    pagination_parameter: PaginationParameterModel | None | Type[Undefined] = Undefined,
) -> FindResultModel:
    client = Backend.client()
    pagination_parameter = pagination_parameter_resolve(pagination_parameter)
    pagination_parameter.sort = "_id"

    tasks = client.get_instances(TEventixTask, pagination_parameter)
    return tasks



import contextlib
import unittest.mock

from pydantic import BaseModel

from eventix.pydantic.event import TEventixEvent


@contextlib.contextmanager
def pytest_eventix_event_posted(event_name: str | None = None):

    class EventMock(BaseModel):
        event: TEventixEvent | None = None

    ret = EventMock()

    with unittest.mock.patch(
        "eventix.functions.eventix_client.EventixClient.ping",
        return_value=True,
    ), unittest.mock.patch(
        "eventix.functions.eventix_client.EventixClient.post_event",
        side_effect=lambda x: [],
    ) as mock_post:
        yield ret
        if not mock_post.called:
            raise Exception("EventixClient.post_event was not called")

        event_obj: TEventixEvent = mock_post.call_args[0][0]
        event = TEventixEvent.model_validate(event_obj.model_dump())
        if event_name is not None and event.name != event_name:
            raise Exception(f"Event name does not match: {event.name}")
        ret.event = event



import logging
from typing import List

from fastapi import APIRouter

from eventix.functions.event import event_post
from eventix.pydantic.event import TEventixEvent
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)

router = APIRouter(tags=["event"])


@router.post("/event")
async def route_event_post(event: TEventixEvent) -> List[TEventixTask]:
    return event_post(event)



import logging
import os
from collections import defaultdict
from typing import Dict, List, Type

import pydash
import yaml

from eventix.functions.task import task_post
from eventix.pydantic.settings import EventixServerSettings


def str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(str, str_presenter)
yaml.representer.SafeRepresenter.add_representer(str, str_presenter)

from eventix.exceptions import TriggerIgnored
from eventix.functions.log import log_prefix, set_log_prefix
from eventix.pydantic.event import TEventixEventTrigger, TEventixEvent
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)


class EventManager(object):
    _triggers: Dict[str, Dict[str, TEventixEventTrigger]] = defaultdict(lambda: dict())

    @classmethod
    def post_event(cls, event: TEventixEvent) -> List[TEventixTask]:
        with set_log_prefix(f"[{event.name}]"):
            log.info(log_prefix("Event posted."))
            triggers: Dict[str, TEventixEventTrigger] = cls._triggers.get(
                event.name, {}
            )
            ret = []
            for trigger in triggers.values():
                with set_log_prefix(f"[{trigger.name}]"):
                    try:
                        task = trigger.execute_code(event)
                        if task.namespace is None or task.namespace =="default":
                            task.namespace = event.namespace
                        task.priority = event.priority
                        task.operator = event.operator

                        updated_task = task_post(task)
                        log.info(
                            log_prefix(
                                f"Trigger executed. Task {updated_task.uid} scheduled."
                            )
                        )
                        ret.append(updated_task)

                    except TriggerIgnored as e:
                        log.info(log_prefix("Trigger ignored:"))

                    except Exception as e:
                        log.error(log_prefix("Trigger failed with error:"))
                        log.error(log_prefix(str(e)))
                        log.exception(e)
        return ret

    @classmethod
    def register_trigger(cls, trigger: TEventixEventTrigger):
        log.info(f'[{trigger.name}] Registered for event "{trigger.event_name}"')
        cls._triggers[trigger.event_name][trigger.name] = trigger


def event_register_trigger(trigger: TEventixEventTrigger):
    return EventManager.register_trigger(trigger)


def event_post(event: TEventixEvent) -> List[TEventixTask]:
    return EventManager.post_event(event)


def event_trigger_yaml_data(trigger: Type[TEventixEventTrigger]) -> dict:
    d = pydash.omit(
        trigger().model_dump(),
        ["created_time", "uid", "updated_time"],
    )
    return d


def event_triggers_yaml_data(triggers: List[Type[TEventixEventTrigger]]) -> dict:
    trigger_data = {
        d["name"]: pydash.omit(d, "name")
        for d in [event_trigger_yaml_data(trigger) for trigger in triggers]
    }
    return trigger_data


def event_triggers_export_yaml(
    out_path: str, triggers: List[Type[TEventixEventTrigger]]
):
    os.makedirs(os.path.dirname(out_path), exist_ok=True)
    with open(out_path, "w") as fp:
        content = dict(triggers=event_triggers_yaml_data(triggers))
        fp.write(yaml.dump(content, indent=2))
    log.info(f"Exported trigger config to '{out_path}'")


def event_triggers_register_from_directory(directory: str):
    for file in os.listdir(directory):
        if file.endswith(".yaml"):
            with open(os.path.join(directory, file)) as fp:
                data = yaml.load(fp, Loader=yaml.SafeLoader)
            triggers = data.get("triggers", {})
            for name, trigger_data in triggers.items():
                trigger_data["name"] = name
                trigger = TEventixEventTrigger.model_validate(trigger_data)
                event_register_trigger(trigger)


def init_triggers():
    settings = EventixServerSettings()
    path = settings.eventix_trigger_config_directory
    if path == "":
        log.info("No trigger config directory specified.")
        return

    log.info(f"Trying to load trigger configs from '{path}'")
    event_triggers_register_from_directory(path)



import logging

from fastapi import APIRouter
from eventix import __version__

log = logging.getLogger(__name__)

router = APIRouter(prefix="")


@router.get("/healthz")
async def router_healthz_get():
    return {"status": "healthy"}


@router.get("/version")
async def router_version_get():
    return {"version": __version__}



import contextlib
# noinspection PyPackageRequirements
import contextvars
import os

from lsidentity.contexts import LsiAccountIdProvider

from eventix.pydantic.task import TEventixTask

namespace_context_var = contextvars.ContextVar('namespace_context_var', default=None)


@contextlib.contextmanager
def namespace_provider(namespace: str):
    # noinspection PyTypeChecker
    token = namespace_context_var.set(namespace)
    yield
    namespace_context_var.reset(token)


@contextlib.contextmanager
def namespace_context():
    # noinspection PyTypeChecker
    namespace = namespace_context_var.get()
    if namespace is None:
        namespace = os.environ.get("EVENTIX_NAMESPACE", "default")
    yield namespace


delay_tasks_context_var = contextvars.ContextVar('delay_tasks_context_var', default=True)


# noinspection PyShadowingNames
@contextlib.contextmanager
def delay_tasks(delay_tasks: bool = True):
    # noinspection PyTypeChecker
    token = delay_tasks_context_var.set(delay_tasks)
    # noinspection PyBroadException
    try:
        yield
    except Exception as e:
        raise e
    finally:
        delay_tasks_context_var.reset(token)

@contextlib.contextmanager
def delay_tasks_context():
    # noinspection PyTypeChecker,PyShadowingNames
    delay_tasks = delay_tasks_context_var.get()
    yield delay_tasks


worker_id_context_var = contextvars.ContextVar('worker_id_context_var', default=None)


@contextlib.contextmanager
def worker_id_provider(worker_id: str):
    # noinspection PyTypeChecker
    token = worker_id_context_var.set(worker_id)
    yield
    worker_id_context_var.reset(token)


@contextlib.contextmanager
def worker_id_context():
    # noinspection PyTypeChecker
    worker_id = worker_id_context_var.get()
    if worker_id is None:
        worker_id = "default"
    yield worker_id


task_priority_context_var: contextvars.ContextVar[int | None] = contextvars.ContextVar(
    'task_priority_context_var',
    default=0
)


@contextlib.contextmanager
def task_priority_provider(priority: int):
    # noinspection PyTypeChecker
    token = task_priority_context_var.set(priority)
    yield
    task_priority_context_var.reset(token)


@contextlib.contextmanager
def task_priority_context():
    # noinspection PyTypeChecker
    priority = task_priority_context_var.get()
    yield priority



task_context_var: contextvars.ContextVar[TEventixTask| None] = contextvars.ContextVar(
    'task_context_var',
    default=None
)


@contextlib.contextmanager
def task_provider(task: TEventixTask):
    # noinspection PyTypeChecker
    token = task_context_var.set(task)
    with LsiAccountIdProvider(task.operator):
        yield task
    task_context_var.reset(token)


@contextlib.contextmanager
def task_context():
    # noinspection PyTypeChecker
    task = task_context_var.get()
    yield task









import logging

from fastapi import APIRouter

from eventix.pydantic.namespaces import NamespacesResponseModel, NamespaceTaskTypesResponseModel

log = logging.getLogger(__name__)

router = APIRouter(tags=["tasks"])


@router.get("/namespaces")
async def router_namespaces_get() -> NamespacesResponseModel:
    raise NotImplementedError()
    # return NamespacesResponseModel(namespaces=[])


@router.get("/namespace/{namespace}/task_types")
async def router_namespace_task_types_get(
    namespace: str,
) -> NamespaceTaskTypesResponseModel:
    raise NotImplementedError()
    # return NamespaceTaskTypesResponseModel(task_types=[])



import dataclasses
from contextlib import contextmanager
from unittest.mock import MagicMock

from lsrestclient import LsRestClient


def mock_eventix_task_scheduler(mocker, return_value=None):
    mock_object = mocker.patch("eventix.functions.task_scheduler.TaskScheduler.task_post", return_value=return_value)
    return mock_object



import datetime
import inspect
import logging
import textwrap
from typing import Callable, Literal, Any, Type

from lsidentity.contexts import LsiAccountId
from pydantic import BaseModel
from pydantic.fields import Field
from pydantic_db_backend_common.pydantic import BackendModel
from pydantic_db_backend_common.utils import utcnow

from eventix.exceptions import TriggerIgnored
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)


class TEventPayload(BaseModel):
    pass


class TEventixEvent(BackendModel):
    namespace: str | None = "default"
    timestamp: datetime.datetime | None = Field(default_factory=utcnow)
    priority: int | None = 0
    payload: dict | None = Field(default_factory=dict)
    name: str | None = None
    operator: str | None = "unknown"

    def __init__(self, **data: Any) -> None:
        if "name" not in data:
            data["name"] = self.__class__.__name__
        if "operator" not in data:
            with LsiAccountId() as operator:
                data["operator"] = operator

        super().__init__(**data)

    def payload_object(self, model: Type[BaseModel]) -> BaseModel:
        return model(**self.payload) if isinstance(self.payload, dict) else self.payload

    def payload_dict(self):
        p = self.payload
        ret = p.model_dump() if isinstance(p, BaseModel) else p
        return ret


TEventixEventTriggerType = Literal["default"]
TEventTriggerExecuteFunction = Callable[[TEventixEvent], TEventixTask]


class TEventixEventTrigger(BackendModel):
    event_name: str
    trigger_type: TEventixEventTriggerType | None = "default"
    namespace: str | None = "default"
    name: str | None = None
    code: str | None = None

    def __init__(self, **data: Any) -> None:
        if "name" not in data:
            data["name"] = self.__class__.__name__
        if "code" not in data:
            data["code"] = self.get_code()

        super().__init__(**data)

    def get_code(self):
        return textwrap.dedent(inspect.getsource(self.execute))

    def ignore(self):
        raise TriggerIgnored(self.name)

    def execute_code(self, event: TEventixEvent):
        code = self.code
        code += "\n\ntask = execute(self, event)"
        f = compile(code, "<string>", "exec")
        exec(f)
        ret: TEventixTask = locals()["task"]
        return ret

    def execute(self, event: TEventixEvent) -> TEventixTask:
        raise NotImplementedError



import contextlib
import functools
from typing import List

from pydantic_db_backend.backend import Backend, has_backend_features

from eventix.exceptions import BackendFeaturesMissing


def backend_features(features: List[str]):
    """
    Raises an error if decorated function is called with an active backend
    not supporting all!!! requested features.

    :param features: list with requested features.
    """
    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            backend = Backend.client()
            if not has_backend_features(backend, features):
                raise BackendFeaturesMissing(backend.__class__.__name__, features)
            else:
                return func(*args, **kwargs)

        return wrapper

    return inner



import json
import os
from collections import OrderedDict
from typing import Optional, List, Dict, Callable

import click
import lsjsonclasses
import pydash
from lsrestclient import LsRestClient
from pydantic import BaseModel
from pydantic_db_backend_common.pagination import pagination_convert_response_list
from pydantic_db_backend_common.pydantic import PaginationResponseModel

from eventix import __version__
from eventix.pydantic.task import TEventixTask


class CLIContext(BaseModel):
    client: LsRestClient
    namespace: Optional[str]

    class Config:
        arbitrary_types_allowed = True


@click.group()
@click.version_option(__version__)
@click.option("-s", "--server", help="server url")
@click.option("-n", "--namespace", help="namespace")
@click.option("--test-client", is_flag=True, default=False)
@click.pass_context
def cli(ctx, server: str = None, namespace: str = None, test_client: bool = False):
    if not test_client:
        LsRestClient(base_url=server, name="eventix")
    ctx.obj = CLIContext(namespace=namespace, client=LsRestClient.client("eventix"))


# noinspection PyUnusedLocal
@cli.group("get")
@click.pass_context
def cli_get(ctx):
    pass


@cli_get.command("tasks")
@click.option("--limit", "limit", default=10, type=int)
@click.option("--skip", "skip", default=0, type=int)
@click.option("--status", "status", default="scheduled", type=str)
@click.pass_context
def cli_get_tasks(ctx, skip: int, limit: int, status: str | None = None):
    namespace = ctx.obj.namespace

    body = dict(skip=skip, limit=limit)

    if status is not None:
        body |= dict(status=status)

    if namespace is not None:
        body |= dict(namespace=namespace)

    # pprint(body)

    r = ctx.obj.client.put("/tasks/by_status", body=body)
    if r.status_code == 200:
        response = PaginationResponseModel.model_validate(r.json())
        tasks = pagination_convert_response_list(response, TEventixTask)

        # pv = r.json()
        # max_result = pydash.default_to(pydash.get(pv, "max_results"), 0)

        start = response.skip + 1
        end = (response.skip * response.limit) + response.limit
        if response.max_results < end:
            end = response.max_results

        pagination_result = f"{start}-{end} of {response.max_results}"

        if len(tasks) != 0:
            entries = [
                OrderedDict(
                    NAMESPACE=task.namespace,
                    UID=task.uid,
                    TASK=task.task,
                    STATUS=task.status,
                    IDENTIFIER="",
                )
                for task in tasks
            ]
            print_table(entries, click.echo)

            click.echo(f"Results {pagination_result}")
        else:
            click.echo(f"No result")

    else:
        raise click.ClickException(r.content)


@cli_get.command("task")
@click.option("--error-only", is_flag=True, default=False)
@click.argument("uid")
@click.pass_context
def cli_get_task(ctx, uid: str, error_only: bool):
    r = ctx.obj.client.get("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        data = r.json()
        status = data.get("status", None)
        if error_only and status in ["retry", "error"]:
            result = pydash.get(data, "result", None)
            if isinstance(result, dict):
                error_class = pydash.get(result, "error_class", None)
            else:
                error_class = None

            click.echo(f"ERROR_CLASS: {error_class}")
            if error_class == "LSoftException":
                error_dict = lsjsonclasses.LSoftJSONDecoder.loads(
                    result["error_message"]
                )
                error = pydash.get(error_dict, "detail.ERROR", None)
                traceback_str = pydash.get(error_dict, "detail.TRACEBACK", None)
                click.echo(f"ERROR:")
                click.echo(error)
                click.echo(f"TRACEBACK:")
                click.echo(traceback_str)
            else:
                print_dict_to_json(result, click.echo)

        else:
            print_dict_to_json(data, click.echo)

    else:
        raise click.ClickException(r.content)


# noinspection PyUnusedLocal
@cli.group("reschedule")
@click.pass_context
def cli_reschedule(ctx):
    pass


@cli_reschedule.command("task")
@click.argument("uid")
# @click.option("--eta") look https://www.markhneedham.com/blog/2019/07/29/python-click-date-parameter-type/
@click.pass_context
def cli_reschedule_task(ctx, uid: str):
    r = ctx.obj.client.get("/task/{uid}/reschedule", params=dict(uid=uid))
    if r.status_code == 200:
        click.echo(f"Rescheduled task {uid}")
    else:
        raise click.ClickException(r.content)


# noinspection PyUnusedLocal
@cli.group("delete")
@click.pass_context
def cli_delete(ctx):
    pass


@cli_delete.command("task")
@click.argument("uid")
@click.pass_context
def cli_delete_task(ctx, uid: str):
    r = ctx.obj.client.delete("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        click.echo(f"Deleted task {uid}")
    else:
        raise click.ClickException(r.content)


@cli.command("dump")
@click.argument("output_directory", required=True)
@click.pass_context
def cli_dump(ctx, output_directory: str):
    os.makedirs(output_directory, exist_ok=True)
    click.echo(f"Dumping to {output_directory}")
    idx = 0
    while True:
        r = ctx.obj.client.put("/tasks/dump", body=dict(limit=10, skip=idx))

        if r.status_code == 200:
            response = PaginationResponseModel.model_validate(r.json())

            if len(response.data) == 0:
                break
            for task in pagination_convert_response_list(response, TEventixTask):
                idx += 1
                click.echo(f"[{idx}] Exporting {task.uid}... ", nl=False)
                fname = os.path.join(output_directory, f"{task.uid}.json")
                # click.echo(f"{fname}...")
                with open(fname, "w", encoding="utf8") as fd:
                    json.dump(
                        task.model_dump(),
                        fd,
                        indent=2,
                        cls=lsjsonclasses.LSoftJSONEncoder,
                    )
                click.echo("done.")
        else:
            raise click.ClickException(r.content)


@cli.command("restore")
@click.argument("input_directory", required=True)
@click.option("--delete-processed", is_flag=True, default=False)
@click.pass_context
def cli_restore(ctx, input_directory: str, delete_processed: bool | None = False):
    click.echo(f"Restoring from {input_directory}")
    filenames = os.listdir(input_directory)
    num_max = len(filenames)
    for idx, filename in enumerate(filenames):
        fpath = os.path.join(input_directory, filename)
        with open(fpath, "r", encoding="utf8") as fd:
            data = json.load(fd, cls=lsjsonclasses.LSoftJSONDecoder)

        click.echo(f"[{idx+1}/{num_max}] Restoring task {data['uid']} ...", nl=False)
        r = ctx.obj.client.put("/task/{uid}", params=dict(uid=data["uid"]), body=data)
        if r.status_code == 200:
            click.echo("done.")
            if delete_processed:
                os.unlink(fpath)
        else:
            raise click.ClickException(r.content)


def print_dict_to_json(data: dict, print_func: Callable = print):
    j = lsjsonclasses.LSoftJSONEncoder.dumps(data, indent=2)
    print_func(j)


def print_table(tdata: List[Dict[str, str]], print_func: Callable = print):
    if len(tdata) == 0:
        return

    keys = list(tdata[0].keys())
    sizes = {}
    lines = []
    for key in keys:
        sizes[key] = max(pydash.concat([len(key)], [len(x[key]) for x in tdata]))

    tdata = pydash.concat([{k: k for k in keys}], tdata)
    for td in tdata:
        line = ""
        for key in keys:
            line += "{text:{width}} ".format(text=td[key], width=sizes[key])
        lines.append(line.strip())

    for line in lines:
        print_func(line)


if __name__ == "__main__":
    cli()



from typing import Annotated, List, Dict, Any

from fastapi import Body
from pydantic import BaseModel


class PaginationParametersModel(BaseModel):
    skip: int | None = 0
    limit: int | None = 25
    sort: List[Dict] | None = None


class PaginationResultModel(PaginationParametersModel):
    data: Any
    max_results: int | None = 0




import logging
import os
import traceback
from typing import Any, Dict, Optional

from fastapi import HTTPException, Request
from pydantic_db_backend.backends.mongodb import MongoDbBackend
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    DispatchFunction,
    RequestResponseEndpoint,
)
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp
from webexception.webexception import WebException

from eventix.pydantic.settings import EventixServerSettings

log = logging.getLogger(__name__)

BACKENDS = {"mongodb": MongoDbBackend}


def init_backend():
    if os.environ.get("APP_CLIENT", "false") != "true":
        settings = EventixServerSettings()
        backend_name = settings.eventix_backend.lower()
        log.info(f"Using {backend_name} backend")
        backend = BACKENDS[backend_name]
        client = backend()
        log.info(client.settings)


# class BackendMiddleware(BaseHTTPMiddleware):
#     def __init__(
#         self,
#         app: ASGIApp,
#         dispatch: DispatchFunction | None = None,
#     ) -> None:
#         super().__init__(app, dispatch)
#
#     async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
#         response = await call_next(request)
#         return response


class ErrorMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        dispatch: DispatchFunction | None = None,
    ) -> None:
        super().__init__(app, dispatch)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            response = await call_next(request)
        except Exception as e:
            return await self.make_exception_response(e)
        return response

    @staticmethod
    async def make_exception_detail(e: Exception) -> Dict:
        tb = traceback.format_exc()
        ret = dict(
            error_class=e.__class__.__name__, error_message=str(e), error_traceback=tb
        )
        return ret

    async def make_exception_response(
        self, e: Exception, headers: Optional[Dict[str, Any]] = None
    ) -> JSONResponse:
        status_code = None
        if isinstance(e, HTTPException):
            status_code = e.status_code
            content = dict(detail=e.detail)

        elif isinstance(e, WebException):
            status_code = e.status_code
            if status_code == 204:
                content = None
            else:
                content = e.dict()
        else:
            if status_code is None:  # no status code found yet
                status_code = 500
            content = dict(detail=await self.make_exception_detail(e))

        if status_code >= 500:
            log.exception(e, stacklevel=2)

        if isinstance(content, dict):
            response = JSONResponse(
                status_code=status_code, content=content, headers=headers
            )
        else:
            response = Response(
                status_code=status_code, headers=headers, content=content
            )
        return response



import contextlib
from typing import List, Type

import pydash

import logging

from eventix.exceptions import EventixHTTPException

log = logging.getLogger(__name__)


@contextlib.contextmanager
def raise_errors(r, exceptions: List[Type[Exception]]):
    exceptions_by_class = {e.__name__: e for e in exceptions}
    if r.status_code < 399:
        yield r
    else:
        try:
            json = r.json()
            detail = pydash.get(json, 'detail', json)
            error_class = pydash.get(detail, 'error_class')
            payload = pydash.get(detail, 'error_payload', {})

            if error_class in exceptions_by_class:
                e = exceptions_by_class[error_class](**payload)
                raise e
            # backend errors
            raise EventixHTTPException(status_code=r.status_code, detail=detail)

        except Exception:
            log.error(r.content)
            raise EventixHTTPException(status_code=r.status_code, detail=r.content)



from __future__ import annotations

import logging
import sys
from typing import Any, List

from lsrestclient import LsRestClient

from eventix.contexts import (
    delay_tasks_context_var,
    namespace_context,
    namespace_context_var,
)
from eventix.exceptions import NoTaskFoundForUniqueKey
from eventix.functions.errors import raise_errors
from eventix.pydantic.event import TEventixEvent
from eventix.pydantic.settings import EventixClientSettings, EventixServerSettings
from eventix.pydantic.task import TEventixTask, EventixTaskStatusEnum

log = logging.getLogger(__name__)


# class EventixClientSession(LsRestClient):
#     def __init__(self, base_url: str = None) -> None:
#         self.client = LsRestClient(base_url, name="")
#         self.base_url = base_url
#         super().__init__()
#
#     def request(
#         self,
#         method,
#         url,
#         *args,
#         **kwargs
#     ) -> Response:  # pragma: no cover
#         return requests.request(
#             method,
#             f"{self.base_url}{url}",
#             *args,
#             **kwargs
#         )


def get_client():
    s = LsRestClient(base_url="nohost://", name="eventix_client")
    # s.headers["Connection"] = "close"
    return s


class EventixClient:
    # interface: Any | None = EventixClientSession()
    interface: Any | None = get_client()
    namespace: str | None = None

    @classmethod
    def set_base_url(cls, base_url):
        if isinstance(cls.interface, LsRestClient):
            log.info(f"Setting EventixClient base_url: {base_url}")
            cls.interface.base_url = base_url

    @classmethod
    def config(cls, config: dict):
        # Be aware that the namespace context is set direct through
        # the context variable.... so no reset possible

        # noinspection PyArgumentList
        settings = EventixClientSettings()
        cls.set_base_url(settings.eventix_url)

        namespace = ""
        if "namespace" in config:
            namespace = config["namespace"]

        if namespace == "":
            namespace = settings.eventix_namespace

        if namespace == "":
            log.error("No EVENTIX_NAMESPACE set.")
            sys.exit()

        namespace_context_var.set(namespace)
        if not settings.eventix_delay_tasks:
            log.info(
                "ATTENTION!!!: Delay tasks disabled. All tasks are run directly. No delay."
            )
        delay_tasks_context_var.set(settings.eventix_delay_tasks)

    @classmethod
    def post_event(cls, event: TEventixEvent) -> List[TEventixTask]:
        with namespace_context() as namespace:
            event.namespace = namespace
            r = cls.interface.post(f"/event", body=event.model_dump())
            with raise_errors(r, []):
                json = r.json()
                assert isinstance(json, list)
                return [TEventixTask(**t) for t in json]

    @classmethod
    def get_task_by_unique_key_and_namespace(cls, unique_key: str, namespace: str = None, stati: list[EventixTaskStatusEnum] = None) -> TEventixTask:

        params = {"stati": [EventixTaskStatusEnum.scheduled.value, EventixTaskStatusEnum.retry.value] if stati is None else stati}

        r = cls.interface.get(f"/task/by_unique_key/{namespace}/{unique_key}", params=params)
        if r.status_code == 404:
            r_data = r.json()
            error_class = r_data.get("error_class")
            if error_class == "NoTaskFoundForUniqueKey":
                raise NoTaskFoundForUniqueKey(unique_key=unique_key, stati=stati)

        with raise_errors(r, []):
            return r

    @classmethod
    def ping(cls) -> bool:
        try:
            r = cls.interface.get(f"/healthz")
            return r.status_code == 200
        except Exception as e:
            log.exception("Error while pinging Eventix server.")
            log.error(e)
            return False



import logging

from eventix.contexts import namespace_context
from eventix.exceptions import backend_exceptions
from eventix.functions.errors import raise_errors
from eventix.functions.eventix_client import EventixClient
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)


class TaskScheduler(EventixClient):

    @classmethod
    def schedule(cls, task: TEventixTask) -> TEventixTask:
        if task.namespace is None:
            with namespace_context() as namespace:
                task.namespace = namespace
        log.debug(
            f"scheduling task: {task.task} "
            f"namespace: {task.namespace} "
            f"uid: {task.uid} "
            f"eta: {task.eta} "
            f"unique_key: {task.unique_key} "
            f"priority: {task.priority}"
        )
        return cls.task_post(task)

    @classmethod
    def task_get(cls, uid: str) -> TEventixTask:
        r = cls.interface.get(f"/task/{uid}")
        with raise_errors(r, backend_exceptions):
            return TEventixTask.model_validate_json(r.content)

    @classmethod
    def task_post(cls, task: TEventixTask) -> TEventixTask:
        r = cls.interface.post("/task", body=task.model_dump())
        with raise_errors(r, backend_exceptions):
            return TEventixTask.model_validate_json(r.content)



from typing import List

from eventix.pydantic.pagination import PaginationResultModel
from eventix.pydantic.task import TEventixTask


class RouterTasksResponseModel(PaginationResultModel):
    data: List[TEventixTask]



import psutil

from eventix.pydantic.task import EventixTaskStatusEnum


class Metrics(object):
    @classmethod
    def dict(cls):
        return {
            "memory": cls.memory_info(),
            "worker": cls.worker_info(),
            "tasks": cls.task_info(),
        }

    @classmethod
    def memory_info(cls):
        memory_info = psutil.Process().memory_info()
        return dict(rss=memory_info.rss, vms=memory_info.vms, data=memory_info.data)

    @classmethod
    def worker_info(cls):
        return {
            "number_of_worker": 0
        }

    @classmethod
    def task_info(cls):
        return {
            EventixTaskStatusEnum.scheduled.value: 0,
            EventixTaskStatusEnum.scheduled.retry: 0,
            EventixTaskStatusEnum.scheduled.error: 0,
            EventixTaskStatusEnum.scheduled.done: 0,
        }


def metrics() -> dict:
    memory_info = psutil.Process().memory_info()
    return {
        "memory": dict(rss=memory_info.rss, vms=memory_info.vms, data=memory_info.data)
    }



from __future__ import annotations

import datetime
import importlib
import logging
import os
import sys
import time
from traceback import format_tb
from typing import List, Dict, Any

import dotenv
import requests.exceptions
from pydantic_db_backend_common.utils import utcnow
from webexception.webexception import WebException

from eventix.contexts import (
    worker_id_context,
    namespace_provider,
    worker_id_provider,
    task_priority_provider, task_provider,
)
from eventix.exceptions import TaskNotRegistered, backend_exceptions
from eventix.functions.core import EventixTaskBase, namespace_context
from eventix.functions.errors import raise_errors
from eventix.functions.eventix_client import EventixClient
from eventix.functions.schedule import schedule_set_next_schedule
from eventix.functions.tools import setup_logging
from eventix.pydantic.schedule import Schedule
from eventix.pydantic.settings import EventixWorkerSettings
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)


class TaskWorker(EventixClient):
    _wait_interval = 10
    _tasks = {}
    _schedule: Dict[str, Schedule] = {}

    namespace: str = None
    worker_id: str = "default"
    schedule_enabled: bool = False
    log_level: str = None

    @classmethod
    def config(cls, config: dict):

        # noinspection PyArgumentList
        settings = EventixWorkerSettings()
        cls.set_base_url(settings.eventix_url)

        cls.config_register_tasks(config)

        if "namespace" in config:
            cls.namespace = config["namespace"]

        if settings.eventix_namespace != "":
            cls.namespace = settings.eventix_namespace

        cls.worker_id = cls.get_worker_id()

        cls.schedule_enabled = settings.eventix_schedule_enabled
        cls.config_schedule(config)

    @classmethod
    def get_worker_id(cls):
        worker_id = os.environ.get("EVENTIX_TASK_WORKER_ID", "default")
        if worker_id != "":
            if worker_id == "K8S_HOSTNAME":
                worker_id = os.environ.get("HOSTNAME", "default").split("-")[-1]
        return worker_id

    @classmethod
    def config_register_tasks(cls, config):
        if "register_tasks" in config:
            cls.register_tasks(config["register_tasks"])

    @classmethod
    def config_schedule(cls, config):
        if "schedule" in config:
            for entry in config["schedule"]:
                s = Schedule.model_validate(entry)
                if s.task not in cls._tasks:
                    raise TaskNotRegistered(s.task)

                schedule_set_next_schedule(s)
                cls._schedule[s.uid] = s
                log.info(f"Scheduled '{s.name}' on '{s.schedule}'")

    @classmethod
    def check_scheduled_tasks(cls):
        log.debug(f"Check scheduled tasks...")
        for s in cls._schedule.values():
            if s.next_schedule <= utcnow():
                # trigger task
                # update schedule
                log.debug(f"triggering scheduled task '{s.name}'")
                cls._tasks[s.task].delay(*s.args, _priority=s.priority, **s.kwargs)
                schedule_set_next_schedule(s)

    @classmethod
    def register_tasks(cls, paths=List[str]):
        log.info("registering tasks...")
        # noinspection PyTypeChecker
        for path in paths:
            try:
                imported_module = importlib.import_module(path)
                for f in filter(
                    lambda x: isinstance(x, EventixTaskBase),
                    [getattr(imported_module, x) for x in dir(imported_module)],
                ):
                    log.info(f"registered '{f.func_name}' from {path}")
                    cls._tasks[f.func_name] = f
            except ImportError as e:
                print(e)

    # with raise_errors(r):
    #     return TaskModel.parse_raw(r.content)

    @classmethod
    def task_next_scheduled(cls) -> TEventixTask | None:
        with namespace_context() as namespace:
            with worker_id_context() as worker_id:
                params = dict(worker_id=worker_id, namespace=namespace)
                r = cls.interface.get(f"/tasks/next_scheduled", params=params)
                try:
                    with raise_errors(r, backend_exceptions):
                        if r.status_code == 200:
                            tm = TEventixTask.model_validate_json(r.content)
                            return tm
                        if r.status_code == 204:
                            return None
                except Exception as e:
                    log.error("Upstream server error:")
                    log.exception(str(e))
                    return None

    @classmethod
    def listen(cls, endless=True, schedule_enabled: bool = False):
        log.info(f"Schedule {'enabled' if schedule_enabled else 'disabled'}")
        log.info("Start listening...")
        while True:
            try:
                if schedule_enabled:
                    cls.check_scheduled_tasks()
                log.debug("Looking for tasks...")
                t = cls.task_next_scheduled()
                if t is not None:
                    cls.execute_task(t)
                else:
                    log.info(f"Nothing to do... waiting {cls._wait_interval}s")
                    if not endless:
                        return
                    time.sleep(cls._wait_interval)
            except requests.exceptions.ConnectionError:
                log.error(
                    f"Upstream Eventix server {cls.interface.base_url} not reachable."
                )
                time.sleep(5)

    @classmethod
    def execute_task(cls, task: TEventixTask):
        try:
            log.debug(f"Executing task: {task.task} uid: {task.uid} ...")

            if task.task not in cls._tasks:
                raise TaskNotRegistered(task=task.task)  # Task not registered in worker

            f = cls._tasks[task.task]
            with task_provider(task):
                with task_priority_provider(task.priority):
                    r = f.run(*task.args, **task.kwargs)
            cls.task_set_result(task, r)

        except TaskNotRegistered as e:
            log.error(f"Error executing task: {task.task} uid: {task.uid}")
            log.exception(e)
            cls.task_set_error(task, TaskNotRegistered(task=task.task))

        except Exception as e:
            log.exception(e)
            cls.task_set_error(task, e)

        finally:
            cls.task_write_back(task)

    @classmethod
    def task_write_back(cls, task: TEventixTask) -> TEventixTask | None:
        try:

            if task.expires is not None and task.expires < utcnow():  # gammel
                # if task is already expired, delete it instead of updating
                cls.interface.delete(f"/task/{task.uid}")
                return None
            else:
                # task not yet expired.... update
                # TODO: use Pydantic v2 json
                r = cls.interface.put(f"/task/{task.uid}", body=task.model_dump())

            with raise_errors(r, backend_exceptions):
                if r.status_code == 200:
                    tm = TEventixTask.model_validate_json(r.content)
                    return tm

        except Exception as e:
            log.error("Exception raised when calling eventix")
            log.exception(e)
            log.error("Exception used this task info")
            log.error(task.model_dump_json())

        return None

    @classmethod
    def load_env(cls):
        dotenv.load_dotenv(".env.local")

    def __init__(self, config: dict) -> None:
        self.load_env()
        self.init_logging()
        self.config(config)

    def init_logging(self):
        worker_id = self.get_worker_id()
        self.log_level = os.environ.get("EVENTIX_WORKER_LOG_LEVEL", "INFO")
        urllib3_logger = logging.getLogger("urllib3.connectionpool")
        urllib3_logger.setLevel(logging.INFO)
        setup_logging(
            level=self.log_level.upper(), module=False, prefix=f"[WORKER-{worker_id}]"
        )

    def start(self, endless: bool | None = True):
        log.info(f"Using namespace: {self.namespace}")
        with namespace_provider(self.namespace):
            with worker_id_provider(self.worker_id):
                self.listen(endless, self.schedule_enabled)

    @staticmethod
    def task_set_error(task: TEventixTask, error: Any):
        if isinstance(error, WebException):
            task.result = error.dict()
        elif isinstance(error, Exception):
            task.result = dict(
                error_class=error.__class__.__name__,
                error_message=str(error),
                error_status_code=500,
                error_traceback=format_tb(error.__traceback__),
                error_payload={},
            )
        else:
            task.result = error

        if task.retry and (task.max_retries is None or task.max_retries != 0):
            task.status = "retry"
            if task.max_retries is not None:
                task.max_retries -= 1  # decrease max_retries until it reaches zero.

            task.eta = utcnow() + datetime.timedelta(seconds=task.error_eta_inc)
            task.error_eta_inc = min([task.error_eta_inc * 2, task.error_eta_max])

            task.worker_id = None

        else:
            task.status = "error"
            if task.error_expires is not None:
                task.expires = utcnow() + datetime.timedelta(seconds=task.error_expires)

        task.worker_expires = None  # removes worker_expires for cleanup prevention

    @staticmethod
    def task_set_result(task: TEventixTask, result: Any):
        task.status = "done"
        if task.store_result:
            task.result = result
            if task.result_expires is not None:
                task.expires = utcnow() + datetime.timedelta(
                    seconds=task.result_expires
                )
        else:
            task.expires = utcnow()

        task.worker_expires = None  # removes worker_expires for cleanup prevention



import os
import sys
from typing import List

from eventix.functions.eventix_client import EventixClient
from eventix.pydantic.task import EventixTaskStatusEnum

try:
    import fastapi
    from fastapi import APIRouter, Body, Query
    from fastapi import Request, Response

except:
    pass


def fastapi_eventix_router_wrapper(self):
    if 'fastapi' not in sys.modules:  # pragma: no cover
        raise Exception('fastapi not installed but required by fastapi_eventix_router_wrapper')

    eventix_router = APIRouter(prefix="", tags=["eventix"])

    @eventix_router.get("/task/by_unique_key/{unique_key}")
    def router_task_by_unique_key_for_namespace_get(unique_key: str, stati: list[EventixTaskStatusEnum] = Query(None)):
        if stati is None:
            stati = [EventixTaskStatusEnum.scheduled.value, EventixTaskStatusEnum.retry.value]
        namespace = os.environ.get("EVENTIX_NAMESPACE", None)
        r = EventixClient.get_task_by_unique_key_and_namespace(unique_key, namespace=namespace, stati=stati)
        return r.json()

    self.include_router(eventix_router)



import logging
from typing import Annotated

from fastapi import APIRouter, Body, Depends
from pydantic_db_backend.fastapi import dep_pagination_parameters
from pydantic_db_backend.pagination import pagination_response
from pydantic_db_backend_common.pydantic import PaginationResponseModel

from eventix.exceptions import NoTaskFound
from eventix.functions.task import (
    task_next_scheduled,
    tasks_by_status,
    tasks_by_task,
    tasks_dump,
)
from eventix.pydantic.pagination import PaginationParametersModel
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)

router = APIRouter(tags=["tasks"])


@router.get("/tasks/next_scheduled")
async def route_tasks_next_scheduled_get(
    worker_id: str, namespace: str
) -> TEventixTask:
    t = task_next_scheduled(worker_id, namespace)
    if t is None:
        raise NoTaskFound(namespace=namespace)
    return t


@router.put("/tasks/by_status")
async def router_tasks_by_status_put(
    status: Annotated[str, Body()] = None,
    namespace: Annotated[str, Body()] = None,
    pagination: PaginationParametersModel = Depends(dep_pagination_parameters),
) -> PaginationResponseModel:
    tasks = tasks_by_status(status=status, namespace=namespace)
    response = pagination_response(tasks)
    return response


@router.put("/tasks/by_task")
async def router_tasks_by_task_put(
    task: Annotated[str, Body()],
    namespace: Annotated[str, Body()] = None,
    pagination: PaginationParametersModel = Depends(dep_pagination_parameters),
) -> PaginationResponseModel:
    tasks = tasks_by_task(task=task, namespace=namespace)
    response = pagination_response(tasks)
    return response


@router.put("/tasks/dump")
async def router_tasks_uids(
    pagination: PaginationParametersModel = Depends(dep_pagination_parameters),
) -> PaginationResponseModel:
    tasks = tasks_dump()
    response = pagination_response(tasks)
    return response



import logging
from contextlib import contextmanager
from contextvars import ContextVar

log = logging.getLogger(__name__)

log_prefix_context_var = ContextVar("log_prefix", default="")


@contextmanager
def set_log_prefix(prefix: str | None = ""):
    existing_prefix = log_prefix_context_var.get()
    token = log_prefix_context_var.set(f"{existing_prefix}{prefix}")
    yield prefix
    log_prefix_context_var.reset(token)


def log_prefix(msg: str) -> str:
    prefix = log_prefix_context_var.get()
    return f"{prefix} {msg}"



from typing import Any, Dict, Optional, List

from pydantic_db_backend_common.exceptions import (
    AlreadyExists,
    NotFound,
    RevisionConflict,
)
from webexception.webexception import WebException


class TaskNotUnique(WebException):
    status_code = 409

    def __init__(self, uid: str) -> None:
        super().__init__(
            f"Task with uid '{uid}' already exists and task unique is not activated for overwriting.",
            uid=uid,
        )


class NoTaskFound(WebException):
    status_code = 204

    def __init__(self, namespace: str) -> None:
        super().__init__(f"No task for namespace {namespace}", namespace=namespace)


class NoTaskFoundForUniqueKey(WebException):
    status_code = 404

    def __init__(self, unique_key: str,stati:str=None) -> None:
        super().__init__(
            f"No task for unique_key {unique_key} and stati {stati} found",
        )
        self.unique_id = unique_key

class TaskUniqueKeyNotUnique(WebException):
    status_code = 409
    def __init__(self, unique_key: str) -> None:
        super().__init__(
            f"Task with unique_key {unique_key} not unique",
        )
        self.unique_id = unique_key

class TaskNotRegistered(WebException):
    status_code = 404

    def __init__(self, task: str) -> None:
        super().__init__(f"Task '{task}' not registered.", task=task)


class EventixHTTPException(WebException):
    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)


class BackendFeaturesMissing(WebException):
    status_code = 501

    def __init__(self, backend_name: str, features: List[str]) -> None:
        super().__init__(
            f"Backend {backend_name} does not support all features {features}", features
        )


class WrongTaskStatus(WebException):
    status_code = 422

    def __init__(self, uid: str, status: str) -> None:
        super().__init__(
            f"Task {uid} has wrong status: '{status}'", uid=uid, status=status
        )


class TriggerIgnored(WebException):
    status_code = 200

    def __init__(self, name: str) -> None:
        super().__init__(f"[{name}] Trigger ignored.")


backend_exceptions = [AlreadyExists, NotFound, RevisionConflict]



import datetime
import logging

from fastapi import APIRouter
from fastapi.params import Query
from pydantic_db_backend.backend import Backend

from eventix.functions.task import task_post, task_by_unique_key, task_reschedule
from eventix.pydantic.task import TEventixTask, EventixTaskStatusEnum

log = logging.getLogger(__name__)

router = APIRouter(tags=["task"])


@router.post("/task")
async def route_task_post(task: TEventixTask) -> TEventixTask:
    return task_post(task)


@router.get("/task/{uid}")
async def route_task_get(uid: str) -> TEventixTask:
    # noinspection PyTypeChecker
    return Backend.client().get_instance(TEventixTask, uid)


@router.delete("/task/{uid}")
async def route_task_delete(uid: str) -> None:
    return Backend.client().delete_uid(TEventixTask, uid)


@router.get("/task/{uid}/reschedule")
async def route_task_reschedule_get(uid: str, eta: datetime.datetime | None = None):
    return task_reschedule(uid, eta)


@router.get("/task/by_unique_key/{unique_key}")
async def router_task_by_unique_key_get(unique_key: str) -> TEventixTask:
    """
    !!!DEPRECATED: use router_task_by_unique_key_for_namespace_get instead

    Fetch a task based on its unique key.

    This asynchronous endpoint retrieves a task identified by
    a given unique key.

    :param unique_key: A string representing the unique key of the task to fetch.
    :return: The task object corresponding to the provided unique key.
    """
    return task_by_unique_key(unique_key=unique_key)

@router.get("/task/by_unique_key/{namespace}/{unique_key}")
async def router_task_by_unique_key_for_namespace_get(
    unique_key: str,
    namespace:str,
    stati:list[EventixTaskStatusEnum] = Query(None)
) -> TEventixTask:
    if stati is None:
        stati = [EventixTaskStatusEnum.scheduled,EventixTaskStatusEnum.retry]
    ret= task_by_unique_key(unique_key=unique_key,namespace=namespace,stati=stati)
    return ret


@router.delete("/task/by_unique_key/{unique_key}")
async def route_task_(unique_key: str) -> None:
    uid = task_by_unique_key(unique_key=unique_key).uid
    return Backend.client().delete_uid(TEventixTask, uid)


@router.put("/task/{uid}")
async def route_task_put(uid: str, task: TEventixTask) -> TEventixTask:
    task.uid = uid  # overwrite uid
    # noinspection PyTypeChecker
    return Backend.client().put_instance(task)



from typing import List

from pydantic import BaseModel


class NamespacesResponseModel(BaseModel):
    namespaces: List[str]


class NamespaceTaskTypesResponseModel(BaseModel):
    task_types: List[str]



import logging
import os

import yaml
from lsrestclient import LsRestClient

from eventix.exceptions import backend_exceptions
from eventix.functions.errors import raise_errors
from eventix.pydantic.relay import RelayModel
from eventix.pydantic.settings import EventixServerSettings
from eventix.pydantic.task import TEventixTask

log = logging.getLogger(__name__)


class RelayManager(object):
    relays: dict = {}

    @classmethod
    def load_config(cls) -> dict | None:
        path = EventixServerSettings().eventix_relay_config
        log.info(f"trying to load relay config at '{path}'")
        if not os.path.exists(path):
            log.info("no relay config found.")
            return None

        with open(path) as fp:
            d = yaml.load(fp, Loader=yaml.SafeLoader)
        config = d.get("relay", None)
        if config is None:
            raise ValueError(
                f"key 'relay' not found in relay configuration at '{path}'"
            )
        return config

    @classmethod
    def add_relay(cls, relay: RelayModel):
        log.info(f"Adding relay for namespace... {relay.namespace} -> {relay.url}")
        cls.relays[relay.namespace] = relay

    @classmethod
    def try_relay(cls, task: TEventixTask) -> TEventixTask | None:
        relay = cls.relays.get(task.namespace, None)
        if relay is None:
            return None
        else:
            log.info(f"Relaying {relay.namespace}/{task.task} to {relay.url}")
            client = LsRestClient(relay.url, name=f"relay-{relay.namespace}")
            r = client.post("/task", body=task.model_dump())
            with raise_errors(r, backend_exceptions):
                return TEventixTask.model_validate_json(r.content)


def init_relay():
    config = RelayManager.load_config()
    if config is not None:
        relays = [RelayModel.model_validate(x) for x in config]
        for relay in relays:
            RelayManager.add_relay(relay)



import os
from typing import Literal

from pydantic_settings import BaseSettings

BackendNames = Literal["mongodb", "couchdb"]


class EventixBaseSettings(BaseSettings):
    eventix_url: str
    eventix_delay_tasks: bool = True
    eventix_namespace: str = ""


class EventixClientSettings(EventixBaseSettings):
    pass


class EventixWorkerSettings(EventixBaseSettings):
    eventix_schedule_enabled: bool = False


class EventixServerSettings(EventixBaseSettings):
    eventix_backend: BackendNames = "mongodb"
    eventix_relay_config: str = os.path.abspath(
        os.path.join(__file__, "../../../relay.yaml")
    )
    eventix_trigger_config_directory: str = ""
    eventix_url: str = ""
    eventix_api_port: int = 8000



import logging

from fastapi import APIRouter

from eventix.functions.metrics import Metrics

log = logging.getLogger(__name__)

router = APIRouter(tags=["metrics"])


@router.get("/metrics")
async def metrics_get():
    return Metrics.dict()
